/// Josh's Logger
/// A Redacted Software Product
/// Developed & Maintained by Josh O'Leary
/// This work is dedicated to the public domain.

/// Special thanks: Maxine Hayes, William J Tomasine II

#pragma once

#include <string>
#include <Event.h>
#include <jlog/color_codes.hpp>

#ifdef __linux__
    #define FUNCTION __PRETTY_FUNCTION__
#endif

#ifdef _WIN32
    #define FUNCTION __FUNCSIG__
#endif

namespace jlog
{
    /// Severity levels for logging. Higher numbers filter less messages out.
    /// @see LOG_LEVEL macro
    enum class severity : uint8_t
    {
        none,       // Show no output
        fatal,      // Show only fatal errors
        error,      // Show fatal and not-necessarily-fatal errors
        warning,    // Show warnings additionally
        verbose,    // Show errors, warnings, and 'verbose' output (i.e. extra information)
        debug,      // Show debugging messages (Basically everything).
    };

    inline severity loglevel = severity::debug; // Default log level always debug

    struct LogEntry
    {
        severity level;
        std::string content;
        std::string timestamp;
    };

    // TODO: Fully implement logging callback.
    static Event<LogEntry> on_log = Event<LogEntry>();

    inline std::vector<LogEntry> log_history;

    /// A single piece of a logging message, with color code, content, and delimiter
    /// These are strung together to build full logger messages in a flexible manner.
    struct token
    {
        colorcode colorCode = color_codes::FG_DEFAULT;
        std::string content = "";
        std::string delimiter = "[]";
    };

    void set_default_logfile(const std::string& filename);

    /// Returns a string timestamp in the format of 'YYYY-MM-DD hh:mm:ss.ms'
    std::string get_timestamp();

    /// Writes an input string directly to standard output
    /// @note Does not append a newline to the message.
    void log_to_console(const std::string& message);
    /// Writes an input string directly to the default destination logfile
    /// @note Does not append a newline to the message.
    /// @see set_default_logfile()
    void log_to_file(const std::string& message);
    /// Writes an input string directly to the specified destination logfile
    /// @note Does not append a newline to the message.
    void log_to_file(const std::string& filename, const std::string& message);
    /// Writes an input string directly to the specified output stream
    /// @note Does not append a newline to the message.
    void log_to_stream(std::ostream stream, const std::string& message);

    /// These are helper functions designed to be wrapped around for easier custom logger building.
    /// @note This file is implemented differently per-platform to handle differences in console color handling.
    /// @see windows/jlog.cpp linux/jlog.cpp
    std::string toks2msg(std::vector<token> tokens, std::function<std::string(token)> formatter);
    std::string consoleMsgFormatter(token t);
    std::string logfileMsgFormatter(token t);
    std::string toks2consoleMsg(std::vector<token> tokens);
    std::string toks2logfileMsg(std::vector<token> tokens);

    /// Parses a sequence of tokens into a printable message, and logs to the following destinations:
    /// standard output (console)
    /// output file (default)
    /// logging callback (Event)
    /// @note This file is implemented differently per-platform to handle differences in console color handling.
    /// @see windows/jlog.cpp linux/jlog.cpp
    void log(const std::vector<token>& tokens, const std::string& filename);
    void log(const std::vector<token>& tokens);
    void ltlog(const std::vector<token>& tokens); // Just for debug purposes

#pragma region Generic Formatters

    /// Returns a pseudo-"stacktrace" formatted sequence of tokens.
    /// @param func The function name/signature to trace back to. Should be provided by a __FUNCTION__ macro variant.
    /// @param file The file name to trace back to. Should be provided by a __FILE__ macro variant.
    /// @param line The source-code line to trace back to. Should be provided by a __LINE__ macro variant.
    std::vector<token> trace_format(
        const std::string& func,
        const std::string& file,
        int line);

    /// Returns a formatted sequence of tokens given a severity and message.
    /// @param severity_name The severity tag to prefix to the message. Could theoretically also be a context.
    /// @param message The actual message to include
    /// @param severity_cc The colorcode to assign to the severity. See color_codes.hpp
    std::vector<token> log_format(
        const std::string& severity_name,
        const std::string& message,
        const colorcode& severity_cc = color_codes::FG_WHITE);

    /// Returns a more detailed formatted sequence of tokens.
    /// @param severity_name The severity tag to prefix to the message. Could theoretically also be a context.
    /// @param message The actual message to include
    /// @param func The function name/signature to trace back to. Should be provided by a __FUNCTION__ macro variant.
    /// @param file The file name to trace back to. Should be provided by a __FILE__ macro variant.
    /// @param line The source-code line to trace back to. Should be provided by a __LINE__ macro variant.
    /// @param severity_cc The colorcode to assign to the severity. See color_codes.hpp
    std::vector<token> log_detailed_format(
    const std::string& severity_name,
    const std::string& message,
    const std::string& func,
    const std::string& file,
    int line,
    const colorcode& severity_cc = color_codes::FG_WHITE);

    /// Returns a token sequence pre-formatted for the INFO log level.
    /// @param message The message to send out.
    std::vector<token> info_format(const std::string& message);

    /// Returns a token sequence pre-formatted for the INFO log level.
    /// @param message The message to send out.
    /// @param func The function name/signature to trace back to. Should be provided by a __FUNCTION__ macro variant.
    /// @param file The file name to trace back to. Should be provided by a __FILE__ macro variant.
    /// @param line The source-code line to trace back to. Should be provided by a __LINE__ macro variant.
    std::vector<token> info_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line);

    /// Returns a token sequence pre-formatted for the WARNING log level.
    /// @param message The message to send out.
    std::vector<token> warning_format(const std::string& message);

    /// Returns a token sequence pre-formatted for the WARNING log level.
    /// @param message The message to send out.
    /// @param func The function name/signature to trace back to. Should be provided by a __FUNCTION__ macro variant.
    /// @param file The file name to trace back to. Should be provided by a __FILE__ macro variant.
    /// @param line The source-code line to trace back to. Should be provided by a __LINE__ macro variant.
    std::vector<token> warning_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line);

    /// Returns a token sequence pre-formatted for the ERROR log level.
    /// @param message The message to send out.
    std::vector<token> error_format(const std::string& message);

    /// Returns a token sequence pre-formatted for the ERROR log level.
    /// @param message The message to send out.
    /// @param func The function name/signature to trace back to. Should be provided by a __FUNCTION__ macro variant.
    /// @param file The file name to trace back to. Should be provided by a __FILE__ macro variant.
    /// @param line The source-code line to trace back to. Should be provided by a __LINE__ macro variant.
    std::vector<token> error_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line);

    /// Returns a token sequence pre-formatted for the FATAL log level.
    /// @param message The message to send out.
    std::vector<token> fatal_format(const std::string& message);

    /// Returns a token sequence pre-formatted for the FATAL log level.
    /// @param message The message to send out.
    /// @param func The function name/signature to trace back to. Should be provided by a __FUNCTION__ macro variant.
    /// @param file The file name to trace back to. Should be provided by a __FILE__ macro variant.
    /// @param line The source-code line to trace back to. Should be provided by a __LINE__ macro variant.
    std::vector<token> fatal_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line);

    /// Returns a token sequence pre-formatted for the VERBOSE log level.
    /// @param message The message to send out.
    std::vector<token> verbose_format(const std::string& message);

    /// Returns a token sequence pre-formatted for the VERBOSE log level.
    /// @param message The message to send out.
    /// @param func The function name/signature to trace back to. Should be provided by a __FUNCTION__ macro variant.
    /// @param file The file name to trace back to. Should be provided by a __FILE__ macro variant.
    /// @param line The source-code line to trace back to. Should be provided by a __LINE__ macro variant.
    std::vector<token> verbose_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line);

    /// Returns a token sequence pre-formatted for the DEBUG log level.
    /// @param message The message to send out.
    std::vector<token> debug_format(const std::string& message);

    /// Returns a token sequence pre-formatted for the DEBUG log level.
    /// @param message The message to send out.
    /// @param func The function name/signature to trace back to. Should be provided by a __FUNCTION__ macro variant.
    /// @param file The file name to trace back to. Should be provided by a __FILE__ macro variant.
    /// @param line The source-code line to trace back to. Should be provided by a __LINE__ macro variant.
    std::vector<token> debug_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line);
#pragma endregion
}

#define INFO(i)    if (jlog::loglevel >= jlog::severity::none) { jlog::log(jlog::info_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

#define VERBOSE(i) if (jlog::loglevel >= jlog::severity::verbose) { jlog::log(jlog::verbose_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

#define DEBUG(i)   if (jlog::loglevel >= jlog::severity::debug) { jlog::log(jlog::debug_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

#define WARNING(i) if (jlog::loglevel >= jlog::severity::warning) { jlog::log(jlog::warning_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

#define ERROR(i)   if (jlog::loglevel >= jlog::severity::error) { jlog::log(jlog::error_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

#define FATAL(i)   if (jlog::loglevel >= jlog::severity::fatal) { jlog::log(jlog::fatal_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

#define LOGLEVEL(i) jlog::loglevel = i;

#define LOGTEST(i) if (jlog::loglevel >= jlog::severity::none) { jlog::ltlog(jlog::info_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

